/*
MPIGLUT: a library lets OpenGL GLUT programs run over MPI.
	http://www.cs.uaf.edu/sw/mpiglut/

This is useful for powerwalls and to solve other distributed rendering problems.


Places where this code runs:
	"backends": MPI machines that do OpenGL rendering.  They've usually got screens directly attached.
	"frontend": User-interface handling machine.  The frontend is usually running a giant virtual DMX server.
	
Pieces of this library:
	mpiglut: simulates GLUT interface for user code.
	dmx: talks with DMX (distributed MultiHead X).
	gui: frontend event-reception code

The MPIglut frontend does not run user code and does not use MPI, but   
does talk with the DMX frontend X server.  The MPIglut
backends all run user code, and use MPI, and display stuff
to their corresponding native backend X server.


Currently for MPIglut, user events like mouse clicks:

	Start at some backend X server, like powerwall0:0

	Get intercepted by the DMX X server, powerwall0:1

	Are fed from the DMX X server to the MPIglut frontend

	Are directly communicated (over a TCP socket) to the MPIglut
	backend "boss" node.

	Are broadcast (via an MPI_Bcast) from the MPIglut boss to
	the MPIglut minions.

	Are passed to user event-handling callbacks by the
	all the MPIglut backends directly and collectively.

So most of the backend X servers are not involved at all in
event handling.  Redisplay events, by contrast, can start in
user code, where they:

	Start with a collective backend call to glutPostRedisplay,
	which is actually ignored except on the backend boss node.

	The backend boss broadcasts the redisplay event (along with
	other events) using MPI_Bcast.

	The backends all then redisplay the screen, by calling the
	user's display callback, (possibly) synchronizing before  
	the buffer swap.

	
There are actually *five* different coordinate systems in use here!

	- global screen coordinates, relative to the frontend display origin.
		This is the virtual coordinate system for user GLUT window calls.
		It's also the common coordinate system used to interchange between all others.
	- local screen coordinates, relative to the backend display origin.
		Used for backend X-window operations.
	- sublocal screen coordinates, where we *can* do drawing.
		Used to divvy up backend displays among several renderer processes.
	
	- window coordinates, relative to the frontend window
		These are the coordinates user drawing and event calls use.
	- subwindow coordinates, relative to the backend subwindow.
		These are the coordinates used for actual rendering.

Normally sublocal <= local <= global, and subwindow <= window (where "<=" means "is a subset of")
and subwindow <= local.  In fact, subwindow == window intersect local.

In the simplest case of a single thread drawing to one fullscreen window, all five
coordinate systems are identical!  In the fullscreen case, window==global and subwindow==sublocal.
For a standard powerwall, global is divided into pieces, but local==sublocal.
For a multi-CPU machine, global==local, but each CPU renders a different sublocal chunk.

Sadly, GLUT has some weird ideas about window coordinate systems.
	Normal drawing happens in OpenGL coordinates, which are viewport-relative.
	Top-level windows are relative to the local screen origin.
	Subwindows are relative to their parent window.

Orion Sky Lawlor, olawlor@acm.org, 2006/10/20 (Public Domain)
*/
#define USE_DMX 1 /* enable DMX distributed X server support? */
#define USE_CUSTOM_FREEGLUT 1 /* enable access to special freeglut functionality? */

#define MPIGLUT_VERSION "0.92 2009-02-04" /* for -mpiglut_version */

#define MPIGLUT_OVERRIDES_GLUT 0 /* turns off mpiglut.h glutX->mpiglutX macro substitutions... */
#include "GL/mpiglut.h"

#if USE_CUSTOM_FREEGLUT
#  include <GL/freeglut_ext.h> /* freeglut extensions-- we need a special version of freeglut 2.4.0! */
#  include <X11/Xlib.h> /* for "Display" */
#  include <X11/Xutil.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdarg.h> /* for va_list stuff in mpiglutTrace */

#ifdef _WIN32
/****** System Specifics: Win32 thread *******/
#include <windows.h>
void porthread_yield(int msec) {
	Sleep(msec);
}

#else 
/******* System Specifics: (non-windows) POSIX thread *******/
#include <unistd.h> /* for usleep */

void porthread_yield(int msec) {
	usleep(msec*1000);
}
#endif

/* You can choose to make all internal mpiglut routines and variables "static" using this #define */
#define INTERNAL /* static */

/* MPI Tags used by mpiglut */
#define MPIGLUT_DMX_INIT_TAG 0xE01Eu
#define MPIGLUT_DMX_PARENT_TAG 0xE01Fu

INTERNAL const char *gui_mpi_name="startup"; /* "frontend" or "backend" */
INTERNAL int trace_mpi_rank=0;/* MPI rank, for tracing purposes */
INTERNAL FILE *trace_file=0; /* trace file */
INTERNAL int trace_verbosity=0;

/* We are this mode and rank--begin tracing. */
INTERNAL void mpiglutTraceBegin(const char *mode,int rank) {
	gui_mpi_name=mode;
	trace_mpi_rank=rank;
}

INTERNAL FILE *mpiglutTraceFile(void) {
	if (trace_file) return trace_file;
	const char *fname_pat="mpiglut_%s%d.trc";
	char buf[1024];
	sprintf(buf,fname_pat,gui_mpi_name,trace_mpi_rank);
	FILE *f=fopen(buf,"a+");
	if (f==0) {
		static int warned=0;
		if (!warned)
			fprintf(stderr,"Could not create trace file %s ",gui_mpi_name);
		warned=1;
		f=stderr;
	}
	trace_file=f;
	return trace_file;
}

/* Debugging trace routine */
void mpiglutTrace(int level,const char *fmt,...)
{
	if (level<=trace_verbosity) 
	{ /* Trace this event to a file. 
		You can watch this file in another terminal with "tail -f mpiglut_frontend0.trc"
	*/
		FILE *f=mpiglutTraceFile();
		if (f==0) return; /* tracing not yet enabled... */
		va_list l;
		va_start(l,fmt);
		vfprintf(f,fmt,l);
		fflush(f);
		va_end(l);
	}
}
#define trace mpiglutTrace

/* Abort utility routine */
void mpiglutAbort(const char *fmt,...)
{
	FILE *f=trace_file?trace_file:stderr;
	fprintf(stderr,"MPIGLUT %s fatal error: ",gui_mpi_name);
	va_list l;
	va_start(l,fmt);
	vfprintf(stderr,fmt,l);
	vfprintf(f,fmt,l);
	va_end(l);
	fflush(stdout);
	fflush(stderr);
	fflush(f);
	fclose(f);
	abort();
}
#define abort mpiglutAbort

/************************ Graphics Utility Classes *******************/
#undef min
#undef max  /* remove stupid windows.h macros */
static int min(int a,int b) {return (a<b)?a:b;}
static int max(int a,int b) {return (a>b)?a:b;}  

/// A 2D integer-coordinates rectangle.
///   Valid pixels go from (lo_x...hi_x-1) x (lo_y...hi_y-1)
typedef struct {
	int lo_x,lo_y,hi_x,hi_y;///< Clip boundaries
} mpiglut_rect;

/// Return a new rectangle with this start and size
mpiglut_rect mpiglut_rect_new(int x,int y,int w,int h)
{
	mpiglut_rect ret;
	ret.lo_x=x;   ret.lo_y=y;
	ret.hi_x=x+w; ret.hi_y=y+h;
	return ret;
}

/// Return this rectangle, shifted by this far.
mpiglut_rect mpiglut_rect_shift(int x,int y,const mpiglut_rect *r)
{
	mpiglut_rect ret;
	ret.lo_x=r->lo_x+x;  ret.lo_y=r->lo_y+y;
	ret.hi_x=r->hi_x+x;  ret.hi_y=r->hi_y+y;
	return ret;
}

/// Return size of rectangle, in pixels
int mpiglut_rect_width(const mpiglut_rect *r) {return r->hi_x-r->lo_x;}
int mpiglut_rect_height(const mpiglut_rect *r) {return r->hi_y-r->lo_y;}

/// Return 1 if this rectangle contains no pixels
int mpiglut_rect_empty(const mpiglut_rect *r) {
	return (r->hi_x<=r->lo_x)||(r->hi_y<=r->lo_y);
}
/// Return 1 if these two rectangles are equal
int mpiglut_rect_equal(const mpiglut_rect *a,const mpiglut_rect *b) {
	return 	(a->lo_x==b->lo_x) &&
		(a->lo_y==b->lo_y) &&
		(a->hi_x==b->hi_x) &&
		(a->hi_y==b->hi_y);
}

/// Return the intersection of these two rectangles
mpiglut_rect mpiglut_rect_intersect(const mpiglut_rect *a,const mpiglut_rect *b) {
	mpiglut_rect ret;
	ret.lo_x=max(a->lo_x,b->lo_x);
	ret.lo_y=max(a->lo_y,b->lo_y);
	ret.hi_x=min(a->hi_x,b->hi_x);
	ret.hi_y=min(a->hi_y,b->hi_y);
	return ret;
}

/// Convert rectangle r from old_coords into new_coords.
///   Both old_coords and new_coords must be themselves expressed in the same coordinate system.
mpiglut_rect mpiglut_rect_coords(const mpiglut_rect *r,const mpiglut_rect *old_coords,const mpiglut_rect *new_coords)
{
	/* Convert r from old_coords to global coordinates */
	mpiglut_rect gr=mpiglut_rect_shift(old_coords->lo_x,old_coords->lo_y,r);
	/* Intersect global-coordinates r with bounds of new coordinate system */
	gr=mpiglut_rect_intersect(&gr,new_coords);
	/* Shift into new coordinate system */
	return mpiglut_rect_shift(-new_coords->lo_x,-new_coords->lo_y,&gr);
}

/// Return true if this point is out of bounds
int mpiglut_rect_oob(int x,int y,const mpiglut_rect *r)
{
	return (x<r->lo_x)||(x>=r->hi_x)||(y<r->lo_y)||(y>=r->hi_y);
}

/* To print out a mpiglut_rect object, use 
	trace(7,"My rect is "RECT_FMT"\n", RECT_ARGS(r));
*/
#define RECT_FMT " (%d,%d  %dx%d)  "
#define RECT_ARGS(r) (r).lo_x,(r).lo_y,mpiglut_rect_width(&(r)),mpiglut_rect_height(&(r))


/***************************** Event Interchange **********************/

/** 
Describes an event sent across the network between 
the mpiglut frontend and backend.
*/
typedef struct {
	int type;
#define MPIGLUT_EVENT_IDLE 10  /**< Nothing's really happening */
#define MPIGLUT_EVENT_MOUSEBUTTON 100  /**< Mouse button going down (value==1) or up (value==0) */
#define MPIGLUT_EVENT_MOUSEMOTION 101  /**< Mouse moving (but buttons not) */
#define MPIGLUT_EVENT_KEYBOARD 200 /**< Keyboard button press (value==1) or release (value==0) */
#define MPIGLUT_EVENT_JOYSTICK 300 /**< Joystick axis move (value==position along axis) or button press (value==1 or 0) */
#define MPIGLUT_EVENT_REPAINT 400 /**< Repaint a window */
#define MPIGLUT_EVENT_WINDOWOPEN 1000 /**< Creating a new window. value==parent window */
#define MPIGLUT_EVENT_WINDOWTITLE 1001 /**< Changing window title: device...end of struct are C-string *chars* (a silly hack!).  value is 0 to set window title; 1 to set icon title. */
#define MPIGLUT_EVENT_TITLE_START_FIELD device
#define MPIGLUT_EVENT_WINDOWVIZ 1005 /**< Changing window visibility: value==0 for show, 1 for hide, 2 for iconify, 3 for permanently destroy */
#define MPIGLUT_EVENT_WINDOWCHANGE 1010 /**< Overall window size or screen position changed */
#define MPIGLUT_EVENT_ENTER 1011 /**< Mouse entering this window */
#define MPIGLUT_EVENT_INIT_SCREENSIZE 2000 /**< At startup: here are the sizes of all the screens */
#define MPIGLUT_EVENT_WINDOW_POINTERS 2001 /**< At window creation: here are the pointers to the DMX subwindows */

	int window; /**< GLUT window number, starting from 1. */
	int value; /**< Varies.  Mouse and keyboard: 0 for released; 1 for pressed.  Joystick: axis value.  Enter event: 1 for entering, 0 for leaving. */
	int device; /**< Device number, starting from 0.  A silly attempt at working with multiple input devices.  For example, the first keyboard is device 0; the first mouse is device 0. */
	int button; /**< Button number.  For mouse, 0 for left, 1 for right, 2 for middle, etc.  For keyboard, either ASCII or else MPIGLUT_SPECIAL_START + a GLUT special key. */
#define MPIGLUT_SPECIAL_START 0x1000000
	int x,y; /**< Position of mouse, for mouse or keyboard events (in global coordinates) */
	int modifiers; /**< State of modifier keys (glutGetModifiers) */
	int mousestate; /**< State of all mouse buttons */
	mpiglut_rect win; /**< Window size, in global coordinates */
	int timestamp; /**< Boss node's backend timer value. */
	int items; /**< Logical number of data items to follow */
	int bytes; /**< Number of bytes of data to follow (ONLY frontend-backend; never in MPI broadcast) */
} mpiglut_event_t;

/* Create a new mpiglut event structure */
void mpiglutMakeEvent(mpiglut_event_t *e,int t) {
	memset(e,0,sizeof(*e));
	e->type=t;
}


/* Here are the routines that interact between various pieces of this program: */

INTERNAL void guiIdleFn(void);
/* Send an event from the GUI frontend to the MPI backends or vice versa */
INTERNAL void guiSendEvent(const mpiglut_event_t *e);
/* Receive an event at the GUI frontend from the MPI backends or vice versa */
INTERNAL int  guiRecvEvent(mpiglut_event_t *e);

/************************ MPI Backend *****************/

/******** Little Utility Code ***********/
/* User event-handling callback functions: */
typedef void (*IdleFn_t)( void );
typedef void (*DisplayFn_t)( void );
typedef void (*ReshapeFn_t)( int w, int h);
typedef void (*MouseFn_t)( int button, int state, int x, int y);
typedef void (*MotionFn_t)( int x, int y);
typedef void (*EntryFn_t)( int state);
typedef void (*KeyboardFn_t)( unsigned char key, int x,int y);
typedef void (*SpecialFn_t)( int key, int x, int y);

/**
 mpiglut_state_t: Describes everything in the current *backend* world setup.
*/
typedef struct {
	int initted; /**< 1 after we've gotten mostly through mpiglutInit */

	MPI_Comm comm; /**< <- MPI Communicator */
	int rank; /**< our 0-based MPI process number */
	int size; /**< size of our MPI communicator--number of ranks */
	int modifiers; /**< glutGetModifiers, as last broadcast from rank 0 */
	int bossRank; /**< rank of the "boss" processor, which broadcasts events */
	
	/** Global coordinates of the entire complete (virtual) frontend screen */
	mpiglut_rect global;
	
	/** Global coordinates of our backend X display */
	mpiglut_rect local;
	
	/** Global coordinates of *our* piece of our local backend (DMX "screen") */
	mpiglut_rect sublocal;

	/* If true, then we need to draw next.  If false, only idle.
	  This prevents continual-redraw applications from denial-of-servicing the underlying GLUT... */
	int drawOrIdle;
	
	/* User's idle function */
	IdleFn_t idle;
	
	/* boss's event queue */
	mpiglut_event_t *Q;
	unsigned int Qin, Qout; /* incoming and outgoing slots in Q array */
#define Qlen 128u /* size of event buffer */

	/* our connection to the frontend display */
#if USE_CUSTOM_FREEGLUT
	Display *display;
#endif
} mpiglut_state_t;

mpiglut_state_t mpiglutState_storage;
/* Look up the mpiglut state */
INTERNAL mpiglut_state_t *mpiglutState(void) {
	if (mpiglutState_storage.initted!=1) {
		mpiglutAbort("Called mpiglut routines without calling mpiglutInit!  Did you #include <mpiglut.h> in your main routine?\n");
	}
	return &mpiglutState_storage;
}


/**
 Describes everything we know about a frontend or backend GLUT window.
 This is virtually 100% event handlers.
*/
typedef struct {
	/* GLUT window ID.  These are assumed to be the same across all processors. */
	int id;
	/* GLUT window ID of parent window.  If zero, this is a top-level window. */
	int parent;
	
	/* If true, this window needs a redraw before the next idle time.  Only read on the boss. */
	int needs_redraw;
	
	/* Global-coordinates frontend window location & size */
	mpiglut_rect window;
	/* Global-coordinates our our subwindow's location & size (our piece of "window") */
	mpiglut_rect subwindow;

	/* Global-coordinates user-requested viewport */
	mpiglut_rect user_viewport;
	/* Global-coordinates OpenGL subwindow viewport */
	mpiglut_rect gl_viewport;
	
	/* User's event callback functions */
	DisplayFn_t display;
	ReshapeFn_t reshape;
	MouseFn_t mouse;
	MotionFn_t motion, passivemotion;
	EntryFn_t entry;
	KeyboardFn_t keyboard, keyboardup;
	SpecialFn_t special, specialup;
/* This ugly little macro calls the function f *iff* f is nonzero. 
   Use it like "USER_CALLBACK(w->reshape)(w,h);"
*/
#define USER_CALLBACK(f) if (f) f

#if USE_CUSTOM_FREEGLUT
	Window xwindow; /* this GLUT window's X11 handle */
#endif
#if USE_DMX
	Window dmx_parent; /* the DMX parent window on the backend screen for this GLUT window */
#endif
} mpiglut_window_t;

/* Create a new window struct */

static mpiglut_window_t **window_slots=0;
static int n_window_slots=0; /* will expand on first use */

/* Look up the mpiglut_window_t with this glut ID */
INTERNAL mpiglut_window_t *mpiglutWindow(int id) {
	int s=id;
	if (s<=0) mpiglutAbort("Current GLUT window %d invalid!",s);
	if (s>=n_window_slots) { /* Expand array to hold value */
		int c,new_n=2*s;
		if (s>1000000) mpiglutAbort("Current GLUT window %d too big!",s);
		window_slots=realloc(window_slots,new_n*sizeof(mpiglut_window_t *));
		for (c=n_window_slots;c<new_n;c++) window_slots[c]=0;
		n_window_slots=new_n;
	}
	if (window_slots[s]==0) {
		mpiglutTrace(4,"mpiglutWindow(w=%d) Creating new window structure\n",s);
		window_slots[s]=(mpiglut_window_t *)malloc(sizeof(mpiglut_window_t));
		memset(window_slots[s],0,sizeof(mpiglut_window_t));
	}
	return window_slots[s];
}

/* Look up the current mpiglut_window_t */
INTERNAL mpiglut_window_t *mpiglutGetWindow(void) {
	return mpiglutWindow(glutGetWindow());
}

INTERNAL void mpiglutDestroyWindowSlot(int window)
{
	free(window_slots[window]);
	window_slots[window]=0;
	mpiglutTrace(4,"mpiglutDestroyWindowSlot(w=%d) Deleted window structure\n",window);
}


/** 
  Push this mpiglut event into the boss's queue.
*/
INTERNAL void mpiglutPushEvent(mpiglut_state_t *s,const mpiglut_event_t *e)
{
	if (s->rank!=s->bossRank) return; /* just ignore events from non-boss processors */
	{
		unsigned int nextIn=(s->Qin+1)%Qlen;
		if (nextIn==s->Qout) mpiglutAbort("mpiglutPushEvent> Event queue full!  (Increase Qlen)\n",0);
		memcpy(&s->Q[s->Qin],e,sizeof(*e));
		s->Qin=nextIn;
	}
}

/** Check for OpenGL errors */
INTERNAL void mpiglutCheck(const char *where) {
	return; /* <- SOAR has OpenGL error I don't even want to know about! */

	mpiglutTrace(10,"mpiglut %s (checking for OpenGL Errors)\n", where);
	GLenum e=glGetError();
	if (e==GL_NO_ERROR) return;
	const GLubyte *errString = gluErrorString(e);
	mpiglutTrace(1,"mpiglut FATAL OpenGL Error in %s: %s\n", where, errString);
	abort("mpiglut FATAL OpenGL Error in %s: %s\n", where, errString);
}


INTERNAL void mpiglutIdleFn(void);

/* Creates the backend */
INTERNAL void guiSpawnBackend(int dmx_screens,int* pargc, char** argv);

/* Handles frontend GUI events.  Never returns */
INTERNAL void guiFrontendRun(int* pargc, char** argv);

/* Sets up connection to GUI for a backend. */
INTERNAL void guiBackendInit(mpiglut_state_t *s,const char *frontend_arg);

#if USE_DMX
/* Return the number of screens on the DMX display, or 0 if not using DMX */
INTERNAL int dmxFrontendInitPre(void);

/* Send DMX screen size to the backends */
INTERNAL void dmxFrontendInitPost(void);
/* Receive DMX screen size from the frontend */
INTERNAL void dmxBackendInit(mpiglut_state_t *s);

INTERNAL void dmxFrontendCreateWindow(int window,int parent,Window w);
INTERNAL Window dmxBackendCreateWindow(mpiglut_state_t *s,int parent,const mpiglut_rect *size);
#endif

INTERNAL int startswith(const char *str,const char *prefix) {
	if (str==0) return 0;
	return 0==strncmp(str,prefix,strlen(prefix));
}

/********** Backend Startup **********/
/* settable command-line options: */

/* Benchmark mode:
	- Make window occupy basically the whole screen
	- Call glutRepaint whenever we go idle.
*/
INTERNAL int mpiglut_benchmark=0;

/* no-sync mode: disables barrier before glutSwapBuffers.
Improves performance, but in the presence of load imbalance
can cause the slower CPUs to fall *way* behind.
*/
INTERNAL int mpiglut_nosync=0;

INTERNAL void trace_args(int v,const char *where,char **argv)
{
	int i;
	mpiglutTrace(v,"Arguments %s:\n",where);
	for (i=0;argv[i];i++)
		mpiglutTrace(v,"	argv[%d]='%s'\n",i,argv[i]);
}

INTERNAL const char *scan_args(int *pargc,char **argv)
{
	const char *frontend_arg=0;
	trace_args(4,"before argument parsing",argv);
	while (argv[1] && startswith(argv[1],"-mpiglut")) {
		if (startswith(argv[1],"-mpiglut_frontend=")) {
			frontend_arg=argv[1];
			mpiglutTrace(1,"Backend will connect to frontend at %s\n",frontend_arg);
		} else if (startswith(argv[1],"-mpiglut_verbose=")) { 	
			sscanf(argv[1],"-mpiglut_verbose=%d",&trace_verbosity);
			mpiglutTrace(1,"Verbosity level set to %d\n",trace_verbosity);
		} else if (0==strcmp(argv[1],"-mpiglut_benchmark")) {
			mpiglutTrace(1,"Benchmark mode enabled\n"); 
			mpiglut_benchmark=1;
		} else if (0==strcmp(argv[1],"-mpiglut_nosync")) {
			mpiglutTrace(1,"Disabling per-frame synchronization\n"); 
			mpiglut_nosync=1;
		} else if (0==strcmp(argv[1],"-mpiglut_version")) { 
			printf("This application compiled with MPIglut version %s\n",
				MPIGLUT_VERSION);
			exit(0);
		} else { /* bad mpiglut argument */
			mpiglutAbort("Invalid mpiglut command-line argument '%s'\n",argv[1]);
		}
		/* Shift arguments down one slot */
		memcpy(&argv[1],&argv[2],(--*pargc - 1)*sizeof(char *));
		argv[*pargc]=0;
	}
	trace_args(4,"after argument parsing",argv);
	return frontend_arg;
}

/** Called by the user.  Calls MPI_Init, sets up the front end,
 and then calls the backend glutInit.
 
 The sequence of operations is a little bit tricky here.
*/
MGAPI void MGAPIENTRY mpiglutInit( int* pargc, char** argv )
{
	mpiglut_state_t *s=&mpiglutState_storage;
	memset(s,0,sizeof(*s));
	mpiglutTrace(20,"Starting mpiglutInit\n");
	
/* Set up MPI.  You've got to do this first, or else the command-line arguments
   aren't there.
*/
	char **margv=argv;
	int err=MPI_Init(pargc,&margv);
	if (err!=0) {
		mpiglutTrace(1,"MPI could not be initialized (error %d)\n",err);
		s->rank=0;
		s->size=1;
		s->bossRank=0;
		s->comm=0;
	}
	else { /* err==0, no problem--MPI initialized. */
		/* Make sure MPI_Finalize gets called before we leave. */
		typedef void (*atexit_fn)(void);
		atexit((atexit_fn)MPI_Finalize);
	
		/* MPICH has the silly habit of *sometimes* changing argv. 
			In particular, argv only changes on *minion* processors, not rank 0.
			FIXME: Is user's argv always long enough to fit all the mpi args?
		*/
		int argc;
		for (argc=0;margv[argc];argc++)
			argv[argc]=margv[argc]; /* copy MPICH's command-line args over to user argv buffer */
		argv[argc]=0; /* nul-terminate argument list */
		*pargc=argc;
	
		s->comm=MPI_COMM_WORLD;
		MPI_Comm_rank(s->comm,&s->rank);
		MPI_Comm_size(s->comm,&s->size);
		s->bossRank=0; // s->size-1; /* choose a boss processor */
		mpiglutTrace(5,"MPI initialized on rank %d--going to barrier\n",s->rank);
		MPI_Barrier(s->comm); /* <- don't start up until everybody's ready */
		mpiglutTrace(5,"Rank %d back from MPI barrier\n",s->rank);
		trace_mpi_rank=s->rank;
	}
	s->initted=1;

/* Check if we're supposed to be a frontend.  
   Backends have the -mpiglut_frontend command-line argument, which tells them
   where to find their frontend.  */
	if (!startswith(argv[1],"-mpiglut_frontend=")) {
		/* then we *are* the frontend */
		if (s->size>1) mpiglutAbort("MPI running, but no -mpiglut_frontend argument!  In this version of mpiglut, the frontend will automatically call mpirun...\n");
	/* Set up tracing */
		system("rm -fr mpiglut_*.trc"); /* remove old trace files */
		mpiglutTraceBegin("frontend",0);

	/* Parse arguments */
		int scratch_argc=*pargc;
		char **scratch_argv=(char **)malloc(sizeof(char *) * scratch_argc);
		memcpy(scratch_argv,argv,sizeof(char *) * scratch_argc);
		scan_args(&scratch_argc, scratch_argv);

	/* Set up DMX and backends */
		int dmx_screens=0;
#if USE_DMX
		dmx_screens=dmxFrontendInitPre();
#endif
		/* We *are* the frontend--need to create some backends */
		guiSpawnBackend(dmx_screens,pargc,argv);
#if USE_DMX
		dmxFrontendInitPost();
#endif
	/* Handle GUI events */
		guiFrontendRun(pargc,argv);
		exit(0);
	}

	/* If we get here, then we're a backend */
	mpiglutTraceBegin("backend",s->rank);
	mpiglutTrace(5,"Rank %d trace setup\n",s->rank);
	
	/* Initially assume global screen is gigantic */
	//s->global=s->local=s->sublocal=mpiglut_rect_new(0,0,-1,-1);
	s->global=s->local=s->sublocal=mpiglut_rect_new(0,0,2000000000,2000000000);
	
	if (0) {
	/* TESTING: Assume local screens consist of small vertical stripes.  Add a small shift, for debugging */
		//int total_wid=1000;
		int wid_per=200; // total_wid/s->size; /* pixels per CPU */
		s->local.lo_x=-200; s->local.lo_y=-100;
		s->sublocal=mpiglut_rect_new(s->rank*wid_per,0,wid_per,1000);
		
	}

/* Parse command-line arguments */
	const char *frontend_arg=scan_args(pargc,argv);
	guiBackendInit(s,frontend_arg);
	
	mpiglutTrace(20,"mpiglutInit calling glutInit with DISPLAY='%s'\n",getenv("DISPLAY"));
	trace_args(15,"before entering glutInit",argv);
	glutInit(pargc,argv);
	
	/* Without DMX, we need to ask GLUT about the screen total size */
	if (s->global.hi_x==-1)
	{
		int w=glutGet(GLUT_SCREEN_WIDTH);
		int h=glutGet(GLUT_SCREEN_HEIGHT);
		s->global=s->local=s->sublocal=mpiglut_rect_new(0,0,w,h);
		printf("No DMX, so reading screen of size %d x %d from GLUT\n",w,h);
	}
		

#if USE_CUSTOM_FREEGLUT
	/* Grab GLUT's X Display using our custom interface */
	s->display=(Display *)glutGetPointer(GLUT_DISPLAY_HANDLE);
#endif

	glutIdleFunc(mpiglutIdleFn);
	trace_args(15,"while leaving mpiglutInit",argv);
	mpiglutTrace(20,"Leaving mpiglutInit\n");
}

/********* Backend Window creation **************/
INTERNAL void mpiglutDisplayFn(void);

/**
  Send off this event to the GUI frontend.  If backendToo is true, 
  also send a local copy of the event to the MPI backends.
*/
INTERNAL void mpiglutSendEvent(mpiglut_event_t *e,int backendToo) {
	mpiglut_state_t *s=mpiglutState();
	if (s->rank!=s->bossRank) return; /* don't mess with events if we're not the boss */
	guiSendEvent(e);
	if (backendToo) mpiglutPushEvent(s,e);
}

/* Forward a copy of window position changes to the frontend */
INTERNAL void mpiglutWindowChange(mpiglut_window_t *w) {
	mpiglut_state_t *s=mpiglutState();
	mpiglut_event_t e; mpiglutMakeEvent(&e,MPIGLUT_EVENT_WINDOWCHANGE);
	e.window=w->id; e.win=w->window;
	w->subwindow=mpiglut_rect_intersect(&w->window,&s->sublocal); /* our portion of our window */
	mpiglutTrace(10,"glutReshape/PositionWindow(w=%d) call: G "RECT_FMT"\n",
			e.window, RECT_ARGS(w->window));
	mpiglutSendEvent(&e,1);
}

/** Forward a copy of window creation and manipulation requests to the front end */
MGAPI int MGAPIENTRY mpiglutCreateWindow( const char* title ) {
	int x=glutGet(GLUT_INIT_WINDOW_X), y=glutGet(GLUT_INIT_WINDOW_Y);
	int wid=glutGet(GLUT_INIT_WINDOW_WIDTH), ht=glutGet(GLUT_INIT_WINDOW_HEIGHT);
	int w;
	if (x==-1) x=50+(rand()%200); /* pick defaults for position-- needed? */
	if (y==-1) y=70+(rand()%200);
	if (mpiglut_benchmark) 
	{ /* benchmark mode--make window huge and fullscreen */
		mpiglut_state_t *s=mpiglutState();
		int inset=10, titlebar=30;
		x=inset; y=inset;
		wid=mpiglut_rect_width(&s->global)-2*inset;
		ht=mpiglut_rect_height(&s->global)-2*inset-titlebar;
		printf("MPIglut: Benchmark window is %d x %d pixels\n",
			wid,ht);
	}
	w=mpiglutCreateSubWindow(0, x,y, wid,ht);
	mpiglutSetWindowTitle(title);
	return w;
}


#if USE_CUSTOM_FREEGLUT
/* This global affects freeglut window creation */
extern Window fgOverrideParentWindow;
#endif

/**
 Create a new window.  
*/
MGAPI int MGAPIENTRY mpiglutCreateSubWindow( int parent, int x, int y, int width, int height ) {
#if USE_DMX
	Window dmx_parent;
#endif
	int window;
	mpiglut_window_t *w=0;
	mpiglut_state_t *s=mpiglutState();
	mpiglut_rect size=mpiglut_rect_new(x,y,width,height); /* parent==0? global coordinates : parent->window coordinates */

/* 1.) Forward a copy of the window creation to the frontend */
	mpiglut_event_t e; mpiglutMakeEvent(&e,MPIGLUT_EVENT_WINDOWOPEN);
	mpiglutCheck("Start of mpiglCreateSubwindow fn");
	mpiglutTrace(3,"%s call starting: parent %d, user-requested size "RECT_FMT"\n",
			parent==0?"glutCreateWindow":"glutCreateSubWindow",
			parent, RECT_ARGS(size));
	e.window=parent; 
	e.win=size;
	mpiglutSendEvent(&e,0);

/* 2.) If needed, set up DMX parent for our subwindow */
#if USE_DMX
	dmx_parent=dmxBackendCreateWindow(s,parent,&size);
#endif
#if USE_DMX && USE_CUSTOM_FREEGLUT
	fgOverrideParentWindow=dmx_parent;
#endif
	
/* 3.) Make a backend "subwindow".
    The actual size will be set in response to a WINDOWCHANGE event, sent by the front end.
*/
	if (parent==0) window=glutCreateWindow("");
	else window=glutCreateSubWindow(parent,0,0,1,1);
	
	glutSetWindow(window);
	glutDisplayFunc(mpiglutDisplayFn); /* connect our display function to new window */

/* 4.) Set up a mpiglut_window structure for this window */
	w=mpiglutWindow(window);
	w->id=window;
	w->parent=parent;
	/* Size will be set in WINDOWCHANGE event */	
#if USE_CUSTOM_FREEGLUT
	w->xwindow=(Window)glutGetPointer(GLUT_WINDOW_HANDLE);
	if (0) { /* Try *hard* to get rid of window decorations.  But still we fail... */
	/* Set the "transient for" hint (works sortof on fvwm, but screws up the window location): */
		XSetTransientForHint(s->display, w->xwindow, w->xwindow);
		
	/* Set the "override redirect" hint (doesn't work on fvwm): */
		if (1) {
			XSetWindowAttributes a;
			a.override_redirect = 1;
			XChangeWindowAttributes(s->display, w->xwindow, CWOverrideRedirect, &a);
		}
		
	/* Set the Motif WM Hint (doesn't work on fvwm) */
		if (1) {
			Atom a = XInternAtom(s->display, "_MOTIF_WM_HINTS", True);
			enum {n=5};
			static const int hints[n]={0,0,0,0,0};
			XChangeProperty(s->display, w->xwindow, a, a, 
				32, /*<- "hints" array is 32-bit quantities */
				PropModeReplace, (unsigned char *) hints,n);
		}
	}
	fgOverrideParentWindow=0; /* stop overriding window parent */
#endif
#if USE_DMX
	w->dmx_parent=dmx_parent;
#endif
#if USE_DMX && USE_CUSTOM_FREEGLUT
	/* Allow X events to percolate up to our parent window, which is DMX.
	  Without this, our backend windows receive mouse events they don't want or need,
	  and DMX doesn't get them.
	*/
	XSelectInput(s->display,w->xwindow,ExposureMask);
#endif

	mpiglutTrace(3,"%s call finished: parent %d, user size "RECT_FMT" -> window %d\n",
			parent==0?"glutCreateWindow":"glutCreateSubWindow",
			parent, RECT_ARGS(size), w->id);
	mpiglutCheck("End of mpiglCreateSubwindow fn");
	return window;
}

/** We store the user's display function */
MGAPI void MGAPIENTRY mpiglutDisplayFunc( void (* callback)( void ) )
	{ mpiglutGetWindow()->display=callback;  }
/** Reshape is called with the size of the entire powerwall's combined screen */
MGAPI void MGAPIENTRY mpiglutReshapeFunc( void (* callback)( int w, int h) )
	{ mpiglutGetWindow()->reshape=callback;  }
MGAPI void MGAPIENTRY mpiglutIdleFunc( void (* callback)( void ) )
	{ mpiglutState()->idle=callback;  }

/** Mouse movement is broadcast from the boss */
MGAPI void MGAPIENTRY mpiglutMouseFunc( void (* callback)( int button, int state, int x, int y) )
	{ mpiglutGetWindow()->mouse=callback; }
MGAPI void MGAPIENTRY mpiglutMotionFunc( void (* callback)( int x, int y) )
	{ mpiglutGetWindow()->motion=callback; }
MGAPI void MGAPIENTRY mpiglutPassiveMotionFunc( void (* callback)( int x, int y) )
	{ mpiglutGetWindow()->passivemotion=callback; }
MGAPI void MGAPIENTRY mpiglutEntryFunc( void (* callback)( int state) )
	{ mpiglutGetWindow()->entry=callback; }

/** Keypresses are broadcast from the boss */
MGAPI void MGAPIENTRY mpiglutKeyboardFunc( void (* callback)( unsigned char key, int x,int y) )
	{ mpiglutGetWindow()->keyboard=callback; }
MGAPI void MGAPIENTRY mpiglutKeyboardUpFunc( void (* callback)( unsigned char key, int x, int y) )
	{ mpiglutGetWindow()->keyboardup=callback; }
MGAPI void MGAPIENTRY mpiglutSpecialFunc( void (* callback)( int key, int x, int y) )
	{ mpiglutGetWindow()->special=callback; }
MGAPI void MGAPIENTRY mpiglutSpecialUpFunc( void (* callback)( int key, int x, int y) )
	{ mpiglutGetWindow()->specialup=callback; }


/********* Backend runtime event handling *********/
INTERNAL void mpiglutHandleEvent(mpiglut_state_t *s,mpiglut_event_t *e);

/** 
  Pull an mpiglut event out of the boss's queue, or return an idle event if none available.
*/
INTERNAL void mpiglutPopEvent(mpiglut_state_t *s,mpiglut_event_t *e)
{
	if (s->rank!=s->bossRank) mpiglutAbort("mpiglutPopEvent> Called on non-boss processor %d!\n",s->rank);
	
	if (s->Qout!=s->Qin) { /* return an item from the local queue */
		unsigned int nextOut=(s->Qout+1)%Qlen;
		memcpy(e,&s->Q[s->Qout],sizeof(*e));
		s->Qout=nextOut;
	} else if (guiRecvEvent(e)) { /* return item from network */
		return;
	} else {
		s->drawOrIdle=!s->drawOrIdle;
		if (s->drawOrIdle) {
			/* Check if any of the windows needs a redraw */
			int w;
			for (w=0;w<n_window_slots;w++) 
				if (window_slots[w]!=0 && window_slots[w]->needs_redraw)
				{ /* fabricate a window redraw event and return it */
					mpiglutMakeEvent(e,MPIGLUT_EVENT_REPAINT);
					e->window=w;
					window_slots[w]->needs_redraw=0;
					return;
				}
		}
		/* Nobody's got anything to do.  We're idle, man. */
		mpiglutMakeEvent(e,MPIGLUT_EVENT_IDLE);
	}
}

/**
  This is the code that runs all the time on all MPI nodes.
*/
INTERNAL void mpiglutIdleFn(void)
{
	//mpiglutCheck("Start of idle fn");
	// mpiglutTrace(20,"- idle -\n");
	mpiglut_state_t *s=mpiglutState();
	mpiglut_event_t e;
	do { /* Grab an event and process it */
		if (s->rank==s->bossRank)
			mpiglutPopEvent(s,&e);
		e.timestamp=glutGet(GLUT_ELAPSED_TIME); /* stamp each before bcast */
		MPI_Bcast(&e,sizeof(e)/sizeof(int),MPI_INT,s->bossRank,s->comm);
		mpiglutHandleEvent(s,&e); /* <- calls user idle */
	} while (e.type!=MPIGLUT_EVENT_IDLE);
	//mpiglutCheck("End of idle fn");
}

/* Last-seen event's timestamp */
INTERNAL int last_timestamp=0;

/** 
  Handle this mpiglut event, which arrived from the network.
  Call the corresponding registered user routine.
*/
INTERNAL void mpiglutHandleEvent(mpiglut_state_t *s,mpiglut_event_t *e)
{
	last_timestamp=e->timestamp;
	if (e->type!=MPIGLUT_EVENT_IDLE) /* don't trace idle events */
		mpiglutTrace(5,"(t=%d): Received MPIGLUT_EVENT %d from network\n",e->timestamp,e->type);
	
	/* Figure out which window the event is for */
	mpiglut_window_t *w=0;
	if (e->window!=0) { /* There's a window associated with this event--make it the current window */
		w=mpiglutWindow(e->window);
		glutSetWindow(w->id);
	}
	if (e->type!=MPIGLUT_EVENT_REPAINT)  /* repaint events don't have modifiers... */
		s->modifiers=e->modifiers;
	
	switch (e->type) {
	case MPIGLUT_EVENT_IDLE:  /**< Nothing's really happening */
	
		if (mpiglut_benchmark>0) { /* repeatedly redisplay, for benchmarking (FIXME: somehow putting this anywhere else screws up resize) */
			mpiglutPostRedisplay();
		}
		if (s->idle) {USER_CALLBACK(s->idle)();}
		else {porthread_yield(10);} /* Sleep for a bit */
		break;
	case MPIGLUT_EVENT_MOUSEBUTTON: /**< Mouse button going up or down */
		mpiglutTrace(10,"MOUSE(w=%d) event: button %d, value %d\n",e->window,e->button,e->value);
		USER_CALLBACK(w->mouse)(e->button,!e->value,e->x,e->y);
		break;
	case MPIGLUT_EVENT_MOUSEMOTION: /**< Mouse moving */
		mpiglutTrace(10,"MOTION(w=%d) event: mouse now at global %d,%d\n",e->window,e->x,e->y);
		if (e->mousestate!=0) {
			USER_CALLBACK(w->motion)(e->x,e->y);
		} else /* no buttons pressed */ {
			USER_CALLBACK(w->passivemotion)(e->x,e->y);
		}
		break;
	case MPIGLUT_EVENT_KEYBOARD: /**< Keyboard button press or release */
		mpiglutTrace(10,"KEYBOARD(w=%d) event: key %d, value %d\n",e->window,e->button,e->value);
		if (e->button>=MPIGLUT_SPECIAL_START) { /* a special key */
			if (e->value==1) {
				USER_CALLBACK(w->special)(e->button-MPIGLUT_SPECIAL_START,e->x,e->y);
			} else { /* key being released */
				USER_CALLBACK(w->specialup)(e->button-MPIGLUT_SPECIAL_START,e->x,e->y);
			}
		} else { /* a normal non-special key */ 
			if (e->value==1) {
				USER_CALLBACK(w->keyboard)((unsigned char)e->button,e->x,e->y);
			} else { /* key being released */
				USER_CALLBACK(w->keyboardup)((unsigned char)e->button,e->x,e->y);
			}
		}
		break;
	case MPIGLUT_EVENT_JOYSTICK: /**< Joystick axis move or button press */
		mpiglutTrace(10,"JOYSTICK(w=%d) event: button %d, value %d\n",e->window,e->button,e->value);
		/* FIXME: handle joysticks */
		break;
	case MPIGLUT_EVENT_REPAINT: {/**< Repaint a window */
		mpiglutTrace(10,"REPAINT(w=%d) event\n",e->window);
		glutSetWindow(e->window);
		mpiglut_window_t *w=mpiglutGetWindow();
		mpiglutTrace(6,"DISPLAY(w=%d) User redisplay function begin\n",e->window);
		USER_CALLBACK(w->display)();
		mpiglutTrace(6,"DISPLAY(w=%d) User redisplay function ended\n",e->window);
		}break;
	case MPIGLUT_EVENT_WINDOWCHANGE: {/**< Overall window size or screen position changed */
		mpiglut_rect win; /* global coordinates of new window */
		w->window=e->win; /* Just accept new global coordinates */
		w->subwindow=mpiglut_rect_intersect(&w->window,&s->sublocal);
		win=w->subwindow;
#if USE_DMX
		if (w->dmx_parent!=0) { /* translate window location so it's relative to our DMX parent window */
			/* HACK: assume DMX parent is in the right place already (suprisingly, this almost works!)
			win.hi_x-=win.lo_x; win.hi_y-=win.lo_y;
			win.lo_x=win.lo_y=0;
			*/
			
			Window root=DefaultRootWindow(s->display);
			Window child_return=0;
			win=mpiglut_rect_coords(&win,&s->global,&s->local); /* window in backend X root coords */
			XTranslateCoordinates(s->display,root,w->dmx_parent,
				win.lo_x,win.lo_y, &win.lo_x,&win.lo_y, &child_return);
			XTranslateCoordinates(s->display,root,w->dmx_parent,
				win.hi_x,win.hi_y, &win.hi_x,&win.hi_y, &child_return);
		
		}
#else
		if (0) { /* DMX not compiled in */ }
#endif	
		else { /* non-DMX case */
			if (w->parent==0) { /* Top-level window: shift from global coordinates to local screen coordinates */
				win=mpiglut_rect_coords(&win,&s->global,&s->local);
			} else { /* Child window: shift from global coordinates to parent's subwindow coordinates */
				win=mpiglut_rect_coords(&win,&s->global,&mpiglutWindow(w->parent)->subwindow);
			}
		}
		
		mpiglutTrace(10,"WINDOWCHANGE(w=%d) event: G "RECT_FMT"   L "RECT_FMT"\n",
			e->window, RECT_ARGS(e->win), RECT_ARGS(win));
		mpiglutCheck("Before moving window");
		if (mpiglut_rect_empty(&win))
		{ /* Local rectangle is empty-- just hide the window */
			glutHideWindow();
		} else { /* Normal case--show window */
			glutReshapeWindow(mpiglut_rect_width(&win),mpiglut_rect_height(&win)); /* *local* width and height */
			glutPositionWindow(win.lo_x,win.lo_y);
			glutShowWindow();
		}
		mpiglutCheck("After moving window");
		glutPostRedisplay();
		USER_CALLBACK(w->reshape)(mpiglut_rect_width(&e->win),mpiglut_rect_height(&e->win)); /* user gets *global* width and height */
		} break;
	case MPIGLUT_EVENT_WINDOWOPEN:
	case MPIGLUT_EVENT_WINDOWVIZ:
		mpiglutAbort("Received invalid window open/close event at backend--these should only be received on the front end GUI!\n",0);
		break;
	case MPIGLUT_EVENT_ENTER: /**< Mouse entering or leaving this window */
		mpiglutTrace(10,"ENTER(w=%d) event: value %d\n",e->window,e->value);
		USER_CALLBACK(w->entry)(e->value);
		break;
	default: mpiglutAbort("Received invalid MPIGLUT_EVENT type=%d from network!\n",e->type);
	};
}

MGAPI int MGAPIENTRY mpiglutGet(GLenum what)
{
	switch (what) {
	case GLUT_SCREEN_WIDTH: return mpiglut_rect_width(&mpiglutState()->global);
	case GLUT_SCREEN_HEIGHT: return mpiglut_rect_height(&mpiglutState()->global);
	
	/* On backends, GLUT only knows about our subwindow, not the entire window */
	case GLUT_WINDOW_WIDTH: return mpiglut_rect_width(&mpiglutGetWindow()->window);
	case GLUT_WINDOW_HEIGHT: return mpiglut_rect_height(&mpiglutGetWindow()->window);
	case GLUT_WINDOW_X: return mpiglutGetWindow()->window.lo_x;
	case GLUT_WINDOW_Y: return mpiglutGetWindow()->window.lo_y;
	
	/* Subtle: GLUT times may drift or quantize differently on the backends;
	  this keeps them locked together.
	*/
	case GLUT_ELAPSED_TIME: return last_timestamp;
	
	default:
		return glutGet(what);
	}
}


/** 
  Handle a glViewport call.
	Input is in window coordinates.
	We secretly convert to local coordinates, and ask OpenGL for only our portion of the viewport.
*/
MGAPI void MGAPIENTRY mpiglViewport(GLint x,GLint y,GLsizei ww,GLsizei wh)
{
	mpiglut_window_t *w=mpiglutGetWindow();
	mpiglut_state_t *s=mpiglutState();
	mpiglut_rect viewport=mpiglut_rect_new(x,y,ww,wh); /* user-requested viewpoint coordinates, relative to window */
	mpiglut_rect gl; /* OpenGL's viewport */
	mpiglutCheck("Start of mpiglViewport fn");
	w->user_viewport=mpiglut_rect_coords(&viewport,&w->window,&s->global); /* to global coordinates */
	mpiglutTrace(10,"glViewport(w=%d) call1: user viewport "RECT_FMT" is G "RECT_FMT"\n",
		w->id,RECT_ARGS(viewport),RECT_ARGS(w->user_viewport));
	
/*
  FIXME: make this just magically work right if the program is setting up an offscreen
     buffer for framebuffer objects, or before trying to do a glCopyTexSubImage2D.
     You might be able to tell the difference by looking at the viewport size--
       a power of two not equal to the window size -> offscreen, so don't modify w->gl.
       not a power of two and/or near-equal to the window size -> onscreen, so fix up w->gl like below.
*/
	if (ww==mpiglut_rect_width(&w->window) && wh==mpiglut_rect_height(&w->window)) 
	{ /* Rendering to an onscreen window */
		/* Can't render outside the boundaries of our subwindow */
		w->gl_viewport=mpiglut_rect_intersect(&w->user_viewport,&w->subwindow);
		mpiglutTrace(4,"glViewport(w=%d) call2: window viewport G"RECT_FMT" converted to OpenGL G"RECT_FMT"\n",
			w->id, RECT_ARGS(w->user_viewport), RECT_ARGS(w->gl_viewport));
		
		/* Pass *subwindow* coordinates to OpenGL */
		gl=mpiglut_rect_coords(&w->gl_viewport,&s->global,&w->subwindow);
	} 
	else 
	{ /* Rendering to an offscreen buffer or something-- don't mess with coordinates */
		w->gl_viewport=w->user_viewport=gl=viewport;
		mpiglutTrace(4,"glViewport(w=%d) call2: user viewport "RECT_FMT" being passed through unmodified\n",
			w->id, RECT_ARGS(viewport));
	}
	
	/* I get "invalid value" errors from OpenGL when called with a negative value */
	if (gl.hi_x<gl.lo_x) gl.hi_x=gl.lo_x; 
	if (gl.hi_y<gl.lo_y) gl.hi_y=gl.lo_y; 
	mpiglutTrace(10,"glViewport(w=%d) call3: gl viewport "RECT_FMT"\n",w->id,RECT_ARGS(gl));
	glViewport(gl.lo_x,gl.lo_y,mpiglut_rect_width(&gl),mpiglut_rect_height(&gl));
	mpiglutCheck("End of mpiglViewport fn");
}

/** Return the coordinates of one of our rectangles.
     value is one of the MPIGLUT_RECT_* types above.
     ret[0] is the left-hand side X coordinate.
     ret[1] is the top Y coordinate (increasing downward).
     ret[2] is the width in pixels.
     ret[3] is the height in pixels.
*/
MGAPI void MGAPIENTRY mpiglutGetRect(int value,int *ret) {
	mpiglut_rect r;
	mpiglut_state_t *s=mpiglutState();
	switch (value) {
	case MPIGLUT_RECT_VIEWPORT: r=mpiglutGetWindow()->gl_viewport; break;
	case MPIGLUT_RECT_GLOBAL: r=s->global; break;
	case MPIGLUT_RECT_LOCAL: r=s->local; break;
	case MPIGLUT_RECT_SUBLOCAL: r=s->sublocal; break;
	default:
		mpiglutAbort("Invalid enumerant %d (%x) passed to mpiglutGetRect\n",value,value);
	};
	ret[0]=r.lo_x;
	ret[1]=r.lo_y;
	ret[2]=r.hi_x-r.lo_x;
	ret[3]=r.hi_y-r.lo_y;
}

/* Return the linear transformation from user clip space pixels [alo,ahi] to 
  OpenGL clip space pixels [blo,bhi].  Idea is that
  	b = scale * a + offset
*/
void mpiglut_make_linear(int alo,int ahi, int blo,int bhi, double *scale,double *offset) {
	/* Step 1: switch from pixels [alo,ahi] to clip space coordinates [-1,1] */
	double cs=2.0/(ahi-alo);
	double co=-1-alo*cs;
	double bl=cs*blo+co, bh=cs*bhi+co; /* b start and end in a's clip space */
	/* we want
		-1 = scale * bl + offset
		+1 = scale * bh + offset
		->  2 = scale * (bh-bl)
	*/
	*scale=2.0/(bh-bl);
	*offset=-1-(*scale)*bl;
}

/** Return the subwindow-from-global-window matrix for the current window and viewport. 
  This matrix returns clip coordinates in our backend's window, given clip coordinates for the virtual global screen.
  
  Our matrix needs to return
     w->gl_viewport clip coordinates   when given   w->user_viewport clip coordinates
  This amounts to a linear transform along the X and Y axes.
*/
MGAPI void MGAPIENTRY mpiglutGetSubwindowMatrixd(double *m) {

	mpiglut_window_t *w=mpiglutGetWindow();
	double xs,xo, ys,yo;
	memset(m,0,sizeof(m[0])*16); /* zero out the matrix */
	m[4*3+3]=m[4*2+2]=1.0; /* *don't* mess with Z or W (FIXME: will W mess this shift up?  Should we compensate for *post*-divide coords?) */
	mpiglut_make_linear(w->user_viewport.lo_x,w->user_viewport.hi_x,
		w->gl_viewport.lo_x,w->gl_viewport.hi_x,   &xs, &xo);
	mpiglut_make_linear(w->user_viewport.lo_y,w->user_viewport.hi_y,
		w->gl_viewport.lo_y,w->gl_viewport.hi_y,   &ys, &yo);
	m[4*0+0]=xs; m[4*3+0]=xo;
	m[4*1+1]=ys; m[4*3+1]=-yo; /* <- negative because OpenGL Y axis up is window system Y axis *down* */
	mpiglutTrace(20,"mpiglutGetSubwindowMatrixd(w=%d) call: x*%f+%f   y*%f+%f\n",
		w->id, xs,xo, ys,yo);
}
MGAPI void MGAPIENTRY mpiglutGetSubwindowMatrixf(float *m) {
	int i; double d[16];
	mpiglutGetSubwindowMatrixd(d);
	for (i=0;i<16;i++) m[i]=d[i];
}


/** Load the subwindow-from-global-window matrix for the current window and viewport. 
  This routine is automatically called by the mpiglut versions of glLoad*, above.
*/
MGAPI void MGAPIENTRY mpiglutLoadSubwindowMatrix( void )
{
	int matmode;
	glGetIntegerv(GL_MATRIX_MODE,&matmode);
	if (matmode==GL_PROJECTION) {
		double m[16];
		mpiglutGetSubwindowMatrixd(m);
		glLoadMatrixd(m);
	}
	else
		glLoadIdentity();
}

/** In GL_PROJECTION mode, these versions  first call glLoadWindowMatrix, then multiply in the user matrix... */
MGAPI void MGAPIENTRY mpiglLoadIdentity( void ) { mpiglutLoadSubwindowMatrix(); }
MGAPI void MGAPIENTRY mpiglLoadMatrixd( const GLdouble *m ) 
	{ mpiglutLoadSubwindowMatrix(); glMultMatrixd(m); }
MGAPI void MGAPIENTRY mpiglLoadMatrixf( const GLfloat *m ) 
	{ mpiglutLoadSubwindowMatrix(); glMultMatrixf(m); }
MGAPI void MGAPIENTRY mpiglLoadTransposeMatrixd( const GLdouble m[16] )  
	{ mpiglutLoadSubwindowMatrix(); glMultTransposeMatrixd(m); }
MGAPI void MGAPIENTRY mpiglLoadTransposeMatrixf( const GLfloat m[16] )
	{ mpiglutLoadSubwindowMatrix(); glMultTransposeMatrixf(m); }

/** Backend hit a redisplay request.  Sadly:
	- You can't just call USER_CALLBACK(w->display)(); here, because
	then the display calls aren't collective anymore and, e.g., barrier will hang.
	- You can't just post a window redisplay, because those are ignored 
	except on the master node.

I think the right thing to do is to either Isend all repaint requests to the master
(with something clever or other to prevent duplicates), or else MPI_Reduce a bitmap
of all repaint requests every frame.

INTERNAL void mpiglutDisplayFn(void)
{
	mpiglut_window_t *w=mpiglutGetWindow();
	USER_CALLBACK(w->display)(); // BAD!  Non-collective display call!
}
*/
INTERNAL void mpiglutDisplayFn(void) {
	mpiglutTrace(10,"REDISPLAY(w=%d) from outside being ignored\n",glutGetWindow());
	
	/* FIXME: handle non-boss nodes here */
	
	/* If we're the boss, send the GLUT display request out to all the nodes... */
	mpiglutPostWindowRedisplay(glutGetWindow());
}

/** On the boss, broadcast redisplay request to all screens */
MGAPI void MGAPIENTRY mpiglutPostWindowRedisplay( int window ) {
	mpiglutTrace(5,"glutPostWindowRedisplay(w=%d)\n",window);
	mpiglut_window_t *w=mpiglutWindow(window);
	w->needs_redraw=1; /* will actually generate redraw event in idle loop */
}


/** On the boss, broadcast redisplay request to all screens */
MGAPI void MGAPIENTRY mpiglutPostRedisplay( void ) {
	mpiglut_state_t *s=mpiglutState();
	if (s->rank!=s->bossRank) return; /* ignore on non-boss CPUs */
	mpiglutTrace(4,"glutPostRedisplay\n");
	int w;
	for (w=0;w<n_window_slots;w++)
		if (window_slots[w]!=0)
			mpiglutPostWindowRedisplay(w);
}

/** Optionally synchronize processors before swapping buffers */
MGAPI void MGAPIENTRY mpiglutSwapBuffers( void ) {
	mpiglut_state_t *s=mpiglutState();
	mpiglutTrace(9,"glutSwapBuffers(w=%d)\n",glutGetWindow());
	mpiglutCheck("Start of mpiglutSwapBuffers");
	if (!mpiglut_nosync) 
	{/* synchronize *before* buffer swap--avoids redraw mismatch */
		glFinish(); /* <- makes sure glutSwapBuffers executes *just* after barrier */
		MPI_Barrier(s->comm); 
	}
	glutSwapBuffers();
	
	mpiglutCheck("End of mpiglutSwapBuffers");
}

/** Return shift, ctrl, alt state broadcast from the boss */
MGAPI int MGAPIENTRY mpiglutGetModifiers(void) 
	{ return mpiglutState()->modifiers; }



MGAPI void MGAPIENTRY mpiglutReshapeWindow( int width, int height )
{
	mpiglut_window_t *w=mpiglutGetWindow();
	w->window.hi_x=w->window.lo_x+width; 
	w->window.hi_y=w->window.lo_y+height;
	mpiglutWindowChange(w);
}
MGAPI void MGAPIENTRY mpiglutPositionWindow( int x, int y )
{
	mpiglut_window_t *w=mpiglutGetWindow();
	w->window.lo_x=x; w->window.lo_y=y;
	mpiglutWindowChange(w);
}

/* Forward a copy of window title changes to the frontend */
INTERNAL void mpiglutWindowTitle(const char *title,int value) {
	mpiglutTrace(10,"%s call: window %d, '%s'\n",
			value==0?"glutSetWindowTitle":"glutSetIconTitle",
			glutGetWindow(), title);
	mpiglut_event_t e; mpiglutMakeEvent(&e,MPIGLUT_EVENT_WINDOWTITLE);
	e.window=glutGetWindow(); e.value=value;
	if ((strlen(title)+1)>=(sizeof(e)-offsetof(mpiglut_event_t,MPIGLUT_EVENT_TITLE_START_FIELD)))
		title="MPIglut window"; /* can't fit long title in "event" struct */
	strcpy((char *)&e.MPIGLUT_EVENT_TITLE_START_FIELD,title);
	mpiglutSendEvent(&e,0);
}

MGAPI void MGAPIENTRY mpiglutSetWindowTitle( const char* title ) {
	mpiglutWindowTitle(title,0);
	glutSetWindowTitle(title);
}
MGAPI void MGAPIENTRY mpiglutSetIconTitle( const char* title ) {
	mpiglutWindowTitle(title,1);
	glutSetIconTitle(title);
}


/* Forward a copy of window visibility changes to the frontend */
INTERNAL void mpiglutWindowViz(int window,int value) {
	mpiglutTrace(10,"glut%sWindow call: window %d, value %d\n",
			value==3?"Destroy":value==2?"Iconify":value==1?"Hide":"Show",window,value);
	mpiglut_event_t e; mpiglutMakeEvent(&e,MPIGLUT_EVENT_WINDOWVIZ);
	e.window=window; e.value=value;
	mpiglutSendEvent(&e,0);
}
MGAPI void MGAPIENTRY mpiglutDestroyWindow( int window ) {
	mpiglutWindowViz(window,3);
	glutDestroyWindow(window);
	mpiglutDestroyWindowSlot(window);
}
MGAPI void MGAPIENTRY mpiglutShowWindow( void ) {
	mpiglutWindowViz(glutGetWindow(),0);
	glutShowWindow();
}
MGAPI void MGAPIENTRY mpiglutHideWindow( void ) {
	mpiglutWindowViz(glutGetWindow(),1);
	glutHideWindow();
}
MGAPI void MGAPIENTRY mpiglutIconifyWindow( void ) {
	mpiglutWindowViz(glutGetWindow(),2);
	glutIconifyWindow();
}

/************** Socket Library (osl/socket.h) ************/
  /*For non-windows (UNIX) systems:*/
#  include <sys/types.h>
#  include <sys/time.h>
#  include <sys/socket.h>
#  include <netinet/in.h>
#  include <arpa/inet.h>
#  include <netdb.h>
#  include <unistd.h>
#  include <fcntl.h>

typedef int skt_ip_t;
#ifndef SOCKET
typedef int SOCKET;
#endif
#ifndef SOCKET_ERROR
#  define SOCKET_ERROR (-1)
#endif
typedef SOCKET SERVER_SOCKET;

INTERNAL struct sockaddr_in skt_build_addr(skt_ip_t IP,int port)
{
	struct sockaddr_in ret={0};
	ret.sin_family=AF_INET;
	ret.sin_port = htons((short)port);
	memcpy(&ret.sin_addr,&IP,sizeof(IP));
	return ret;  
}

skt_ip_t _skt_invalid_ip=0;

#ifdef linux /* needed for address determination code below... */
#include <ifaddrs.h> /* for getifaddrs */
#include <net/if.h> /* for IFF_RUNNING */
#endif

/* Create a TCP server socket, optionally listening on this port and IP*/
INTERNAL SERVER_SOCKET skt_server_ip(unsigned int *port,skt_ip_t *ip)
{
	SERVER_SOCKET             ret;
	int on = 1; /* for setsockopt */
	int connPort=(port==NULL)?0:*port;
	struct sockaddr_in addr=skt_build_addr((ip==NULL)?_skt_invalid_ip:*ip,connPort);
	socklen_t          len=sizeof(addr);

	ret = socket(PF_INET, SOCK_STREAM, 0);
	if (ret == SOCKET_ERROR) mpiglutAbort("Error creating server socket\n");
	/* Prevents 3-minute socket reuse timeout after a server crash. */
	setsockopt(ret, SOL_SOCKET, SO_REUSEADDR, (const char *)&on, sizeof(on));

	if (bind(ret, (struct sockaddr *)&addr, sizeof(addr)) == SOCKET_ERROR) 
	  mpiglutAbort("Error binding server socket.  Is another process listening on that port already?");
	listen(ret,5);
	if (getsockname(ret, (struct sockaddr *)&addr, &len) == SOCKET_ERROR) 
	  mpiglutAbort("Error getting name on server socket.");

	if (port!=NULL) *port = (int)ntohs(addr.sin_port);
	if (ip!=NULL) {
		memcpy(ip, &addr.sin_addr, sizeof(*ip));
#ifdef linux
		if (*ip == _skt_invalid_ip) 
		{ /* drat--it's not easy to find our local IP! */
			/* Code snippet from  Jens Alfke
			http://lists.apple.com/archives/macnetworkprog/2008/May/msg00013.html */
			struct ifaddrs *interfaces=0;
			if( getifaddrs(&interfaces) == 0 ) {
				struct ifaddrs *interface;
				for( interface=interfaces; interface; interface=interface->ifa_next ) {
					if( (interface->ifa_flags & IFF_UP) && ! (interface->ifa_flags & IFF_LOOPBACK) ) {
						const struct sockaddr_in *addr = (const struct sockaddr_in*) interface->ifa_addr;
						if( addr && addr->sin_family==AF_INET ) {
							memcpy(ip, &addr->sin_addr, sizeof(*ip));
							break;
						}
					}
				}
				freeifaddrs(interfaces);
			}
		}
#endif
	}
	return ret;
}

/* Accept an incomming connection on this server socket */
INTERNAL SOCKET skt_accept(SERVER_SOCKET src_fd,skt_ip_t *pip, unsigned int *port)
{
	socklen_t len;
	struct sockaddr_in addr={0};
	SOCKET ret = accept(src_fd, (struct sockaddr *)&addr, &len);
	if (ret == SOCKET_ERROR) mpiglutAbort("skt_accept: Error in client connection\n");
	if (port!=NULL) *port=ntohs(addr.sin_port);
	if (pip!=NULL) memcpy(pip,&addr.sin_addr,sizeof(*pip));
	return ret;
}

/* Connect to this TCP address */
INTERNAL SOCKET skt_connect(skt_ip_t ip, int port, int timeout)
{
	struct sockaddr_in addr=skt_build_addr(ip,port);
	SOCKET ret=socket(AF_INET, SOCK_STREAM, 0);
	int err = connect(ret, (struct sockaddr *)&(addr), sizeof(addr));
	if (err == SOCKET_ERROR) mpiglutAbort("skt_connect: Could not connect to port %d\n",port);
	return ret;
}


/* Query if this socket is readable.  Waits up to msec long.  
   Returns 0 if timeout expires before readable. */
INTERNAL int skt_select1(SOCKET fd,int msec)
{
	int sec=msec/1000;
	fd_set  rfds;
	struct timeval tmo;
	int  secLeft=sec;
	int  begin=0, nreadable;

	FD_ZERO(&rfds);
	FD_SET(fd, &rfds);

	tmo.tv_sec=msec/1000;
	tmo.tv_usec = msec*1000;
	nreadable = select(1+fd, &rfds, NULL, NULL, &tmo);
	return nreadable>0;
}


/**************** Frontend/Backend Event Transmission **************
  FIXME: this is all unix-specific.  It should be possible to put the frontend on a remote network box. 
*/
#include <unistd.h>
#include <sys/socket.h>

INTERNAL int gui_mpi_socket; /* gui-to-from-mpi communication socket */

INTERNAL int mpiglutFrontendWindows=0;

/* Creates the backend */
INTERNAL void guiSpawnBackend(int dmx_screens,int* pargc, char** argv)
{
	mpiglutTrace(10,"guiSpawnBackend starting\n");
	unsigned int port=0;
	skt_ip_t ip=_skt_invalid_ip;
	SERVER_SOCKET gui_serv=skt_server_ip(&port,&ip);
	char frontend_arg[1000];
	sprintf(frontend_arg,"-mpiglut_frontend=%08x:%d",ip,port);
	
	//if (fork()==0) 
	{ /* Set up the new backend process */
		int argc=*pargc;
		char **bargv=(char **)malloc(sizeof(char *)*(6+argc));
		argv[argc]=0;
		if (dmx_screens==0) 
		{ /* Just call ourselves */
			bargv[0]=argv[0]; /* program name */
			bargv[1]=frontend_arg;
			memcpy(&bargv[2],&argv[1],argc*sizeof(char *));
		} else /* dmx_screens>0 */ {
		/*  There are many screens to talk to, so run
			mpirun -np <backend nodes> <user program and args> 
		*/
			bargv[0]="mpirun";
			if (getenv("MPIRUN")) bargv[0]=getenv("MPIRUN");
			bargv[1]="-np";
			char bnodes[100];
			sprintf(bnodes,"%d",dmx_screens);
			bargv[2]=bnodes;
			bargv[3]=argv[0]; /* program name */
			bargv[4]=frontend_arg;
			memcpy(&bargv[5],&argv[1],argc*sizeof(char *));
		}
		trace_args(8,"Frontend spawning backend",bargv);
		if (fork()==0) 
		{ /* We are the new backend process-- run it. */
			execvp(bargv[0],bargv);
			perror("execvp");
			mpiglutAbort("Error running mpirun %s %s %s %s!",
				bargv[1],bargv[2],bargv[3],bargv[4]);
		}
	}
	/* else I am the parent frontend process */
	
	/* Listen for the backend's connection */
	mpiglutTrace(1,"Waiting for backend connection to our '%s'\n",frontend_arg);
	int i;
	for (i=0;i<20;i++) {
		if (skt_select1(gui_serv,1000)>0) break;
		else printf("Still waiting for backend to connect to '%s'\n",frontend_arg);
	}
	
	gui_mpi_socket=skt_accept(gui_serv,0,0);
	mpiglutTrace(1,"Backend connected on socket %d\n",gui_mpi_socket);
}

/* Set up a GUI backend */
INTERNAL void guiBackendInit(mpiglut_state_t *s,const char *frontend_arg) {
	mpiglutTrace(10,"guiBackendInit running on rank %d (frontend=%s)\n",s->rank,frontend_arg);
	
	if (frontend_arg) {
		/* Frontend is talking with DMX, somewhere out there on the net */
		if (s->rank==s->bossRank) 
		{ /* Connect to the frontend's socket */
			int port; skt_ip_t ip;
			if (2!=sscanf(frontend_arg,"-mpiglut_frontend=%08x:%d",&ip,&port)) 
				mpiglutAbort("Can't parse mpiglut_frontend argument '%s'\n",frontend_arg);
			mpiglutTrace(10,"guiBackendInit boss connecting to front end IP=%08x and port=%d\n",ip,port);
			gui_mpi_socket=skt_connect(ip,port,60);
			mpiglutTrace(10,"guiBackendInit boss connected, socket=%d\n",gui_mpi_socket);
			s->Q=(mpiglut_event_t *)malloc(sizeof(mpiglut_event_t)*Qlen);
			s->Qin=s->Qout=0;
		}
		
#if USE_DMX /* If there's a frontend, we must call dmxInit to set up our DISPLAY variable */
		dmxBackendInit(s);
#endif
	} 
}

/** Send an event from the GUI frontend to the MPI backends or vice versa */
INTERNAL void guiSendEvent(const mpiglut_event_t *e) {
	if (gui_mpi_socket<=0) mpiglutAbort("Called guiSendEvent without a socket!");
	mpiglutTrace(20,"guiSendEvent type %d\n",e->type);
	if (sizeof(*e)!=send(gui_mpi_socket,e,sizeof(*e),0))
		mpiglutAbort("guiSendEvent> Error sending data");
}
/** Receive an event at the GUI frontend from the MPI backends or vice versa  */
INTERNAL int guiRecvEvent(mpiglut_event_t *e) {
	if (gui_mpi_socket<=0) mpiglutAbort("Called guiRecvEvent without a socket!");
	if (skt_select1(gui_mpi_socket,0)) {
		if (sizeof(*e)!=recv(gui_mpi_socket,e,sizeof(*e),0))
			mpiglutAbort("guiRecvEvent> Error receiving data");
		mpiglutTrace(20,"guiRecvEvent type %d\n",e->type);
		return 1; /* real event */
	}
	else { /* no network data */
		return 0; /* no event */
	}
}


/** Send this block of binary data from the GUI to MPI or vise versa */
INTERNAL void guiSendData(const void *data,int nBytes, const char *where) {
	mpiglutTrace(20,"guiSendData %p  %d bytes: %s\n",data,nBytes,where);
	if (gui_mpi_socket<=0) mpiglutAbort("Called guiSendData without a socket!");
	if (nBytes != send(gui_mpi_socket,data,nBytes,0))
		mpiglutAbort("guiSendData> Error sending data");
	mpiglutTrace(20,"guiSendData complete\n");
}
/** Receive this block of binary data from the GUI to MPI or vise versa */
INTERNAL void guiRecvData(void *data,int nBytes, const char *where) {
	mpiglutTrace(20,"guiRecvData %p  %d bytes: %s\n",data,nBytes,where);
	if (gui_mpi_socket<=0) mpiglutAbort("Called guiRecvData without a socket!");
	if (nBytes != recv(gui_mpi_socket,data,nBytes,0))
		mpiglutAbort("guiRecvData> Error receiving data");
	mpiglutTrace(20,"guiRecvData %p  %d bytes complete\n",data,nBytes);
}

/** Send a synchronization tag */
INTERNAL void guiSendSync(int tag) { guiSendData(&tag,sizeof(tag),"Synchronization tag"); }
/** Receive and check a synchronization tag */
INTERNAL void guiRecvSync(int expected_tag) {
	int tag; 
again:
	mpiglutTrace(3,"guiRecvSync> Expecting synchronization tag %d (0x%x) from frontend\n",expected_tag,expected_tag);
	guiRecvData(&tag,sizeof(tag),"Synchronization tag"); 
	if (tag!=expected_tag) 
	{ // It's possible to receive normal events from the frontend when expecting something else...
		
		mpiglut_event_t e;
		e.type=tag;
		mpiglutTrace(3,"guiRecvSync> Handling reentrant event type %d (0x%x) {\n",tag,tag);
		recv(gui_mpi_socket,&e.window,sizeof(e)-sizeof(e.type),0);
		mpiglutPushEvent(mpiglutState(),&e); // Queue up event in boss's queue...
		mpiglutTrace(3,"guiRecvSync> } done with reentrant event handling inside guiRecvSync\n");
		goto again;
	}
}


/********************************* (subs) Subscreen Code ************************
 Interaction with DMX goes here.
 
*/
#if USE_DMX
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/dmxext.h>

/* Variables used only on the frontend */
INTERNAL Display *dmx_display; /* backends do NOT talk directly to DMX, so this is NULL */

/* Variables used by both frontend and backend */
INTERNAL int dmx_screen_count; /* number of subscreens managed by DMX */
INTERNAL int dmx_enabled; /* 1 if DMX exists; 0 for no DMX */

/* Struct passed from frontend to backend at startup in DMX mode */
typedef struct {
	int rank; /* Target backend rank (0..size-1) */
	mpiglut_rect global, local, sublocal; /* rectangles for this backend to use (see mpiglut_state) */
	char display[64]; /* X DISPLAY variable for this backend to use (normally local) */
} mpiglut_init_info;

/** Return the number of backend screens, or 0 if no backend is available. */
INTERNAL int dmxFrontendInitPre(void) 
{
	int evt,err;
	const char *dpyName=getenv("DISPLAY");
	trace(10,"dmxFrontendInit starting on DISPLAY '%s'\n",dpyName);
	dmx_display=XOpenDisplay(dpyName);
	if (dmx_display==NULL) /* bad DISPLAY variable */
		abort("Cannot open DISPLAY '%s'!",dpyName);
	
	if (!DMXQueryExtension(dmx_display,&evt,&err)) {
		// Crap.  This display isn't DMX-compatible.
		trace(1,"DMX does not exist on DISPLAY '%s'\n",dpyName);
		dmx_enabled=0;
		dmx_screen_count=-1;
		trace(10,"dmxFrontendInit complete (no DMX. crap)\n");
		return 0;
	} 
	// else this display uses DMX...
	dmx_enabled=1;
	DMXGetScreenCount(dmx_display,&dmx_screen_count);
	return dmx_screen_count;
}

/** Start talking to the backend screens */
INTERNAL void dmxFrontendInitPost(void)
{
	int s;
	trace(10,"DMX: %d screens\n",dmx_screen_count);
	guiSendSync(MPIGLUT_DMX_INIT_TAG);
	guiSendData(&dmx_screen_count,sizeof(dmx_screen_count),"dmxFrontendInit screen count"); /* send off screen count to backend boss */
	
	for (s=0;s<dmx_screen_count;s++) {
		mpiglut_init_info init;
		DMXScreenAttributes attr; /* per-screen attributes (size, location, etc.) */
		DMXGetScreenAttributes(dmx_display,s,&attr);
		trace(15,"DMX screen %d: \n",s);
		trace(15,"	display: %s (logical %d)\n",attr.displayName,attr.logicalScreen);
		trace(15,"	screen: %dx%d at %d,%d\n",
			attr.screenWindowWidth,attr.screenWindowHeight,
			attr.screenWindowXoffset,attr.screenWindowYoffset);
		trace(15,"	root: %dx%d at %d,%d  (global %d,%d)\n",
			attr.rootWindowWidth,attr.rootWindowHeight,
			attr.rootWindowXoffset,attr.rootWindowYoffset,
			attr.rootWindowXorigin,attr.rootWindowYorigin);
		/** Compute this screen's rectangle in global coordinates */
		init.rank=s;
		sprintf(init.display,"%s",attr.displayName); // ,attr.logicalScreen);
		init.global=mpiglut_rect_new(0,0,
			DisplayWidth (dmx_display,DefaultScreen(dmx_display)),
			DisplayHeight(dmx_display,DefaultScreen(dmx_display))
		);
		
		/* Local coordinates: the backend X server's native coordinate system */
		init.local=mpiglut_rect_new(
			attr.rootWindowXorigin-attr.screenWindowXoffset, 
			attr.rootWindowYorigin-attr.screenWindowYoffset, 
			attr.rootWindowWidth+attr.screenWindowXoffset, 
			attr.rootWindowHeight+attr.screenWindowYoffset
		);
		
		/* Actual DMX "screen"-- the geometric area we should manage */
		init.sublocal=mpiglut_rect_new(
			attr.rootWindowXorigin,attr.rootWindowYorigin,
			attr.rootWindowWidth,attr.rootWindowHeight);
		
		guiSendData(&init,sizeof(init),"dmxFrontendInit dmx init data");
	}
	guiSendSync(MPIGLUT_DMX_INIT_TAG);
	trace(10,"dmxFrontendInit complete\n");
}
INTERNAL void dmxBackendInit(mpiglut_state_t *s)
{
	int nScreens;
	mpiglut_init_info init; /* attributes of my screen */
	trace(10,"dmxBackendInit starting, originally on DISPLAY '%s'\n",getenv("DISPLAY"));

/* Get screen count */
	if (s->rank==s->bossRank) { /* boss processor-- receive count from front end */
		guiRecvSync(MPIGLUT_DMX_INIT_TAG);
		guiRecvData(&nScreens,sizeof(nScreens),"dmxBackendInit screen count");
	}
	MPI_Bcast(&nScreens,1,MPI_INT, s->bossRank,s->comm);
	if (nScreens==-1) {/* DMX not enabled-- skip this */
		dmx_enabled=0;
		guiRecvSync(MPIGLUT_DMX_INIT_TAG);
		trace(10,"dmxBackendInit returning early-- DMX not enabled\n");
		return;
	}
	// else DMX is enabled
	dmx_screen_count=nScreens;
	dmx_enabled=1;

/* Get screen data */
	if (s->rank==s->bossRank) 
	{ /* boss processor-- receive info from front end and distribute to minions */
		int scr;
		if (dmx_screen_count!=s->size)
			printf("WARNING: There are %d DMX screens, but %d MPI processes.  You need to run this program like 'mpirun -np %d ...'\n",
				dmx_screen_count,s->size,dmx_screen_count);
		
		for (scr=0;scr<dmx_screen_count;scr++) {
			mpiglut_init_info cur;
			guiRecvData(&cur,sizeof(cur),"dmxBackendInit backend init data");
			if (scr==s->rank) /* these are mine */
				init=cur; 
			else if (scr<s->size) /* those are one of our minion's */
				MPI_Send(&cur,sizeof(cur),MPI_BYTE,
					scr, MPIGLUT_DMX_INIT_TAG, s->comm);
			else { /* unused screen with no backend-- skip it! */
			}
		}
		guiRecvSync(MPIGLUT_DMX_INIT_TAG);
	}
	else { /* minion processor-- receive our init info from our boss */
		MPI_Status sts;
		MPI_Recv(&init,sizeof(init),MPI_BYTE,
			s->bossRank, MPIGLUT_DMX_INIT_TAG, s->comm,&sts);
	}
	
/* Backend startup */
	if (init.rank!=s->rank) mpiglutAbort("Logic error in dmxBackendInit: rank %d got init data for %d\n",s->rank,init.rank);
	
	s->global=init.global;
	s->local=init.local;
	s->sublocal=init.sublocal;
	setenv("DISPLAY",init.display,1);
	
	trace(15,"Backend rank %d of %d: \n",s->rank,s->size);
	trace(15,"	display: %s\n",init.display);
	trace(15,"	global: G "RECT_FMT"\n"
		"	local: G "RECT_FMT"\n"
		"	sublocal: G "RECT_FMT"\n",
		RECT_ARGS(s->global), RECT_ARGS(s->local), RECT_ARGS(s->sublocal)
	);
	
	trace(10,"dmxBackendInit complete\n");
}

INTERNAL void dmxFrontendCreateWindow(int window,int parent,Window win)
{
	int w,wan=dmx_screen_count;// Length of wa array
	DMXWindowAttributes *wa;
	if (!dmx_enabled) return;
	trace(10,"dmxFrontendCreateWindow starting (w=%d,parent=%p,x Window %p)\n",window,parent,win);
	/* per-backend window attributes */
	wa=(DMXWindowAttributes *)malloc(sizeof(DMXWindowAttributes)*dmx_screen_count); 
	
	// DMXForceWindowCreation(dmx_display,win); //<- seems like it'd be needed, but this just hangs!
	
	do {
		DMXSync(dmx_display);
		DMXGetWindowAttributes(dmx_display,win,&wan,dmx_screen_count,wa);
		trace(15,"Window information available for %d of %d screens\n",wan,dmx_screen_count);fflush(stdout);
		for (w=0;w<wan;w++) {
			trace(15,"   Window %d: screen %d, Xwin %p, pos=(tl %d,%d size %d,%d), viz=(tl %d,%d size %d,%d)\n",
				w,wa[w].screen,wa[w].window,
				wa[w].pos.x,wa[w].pos.y,wa[w].pos.width,wa[w].pos.height,
				wa[w].vis.x,wa[w].vis.y,wa[w].vis.width,wa[w].vis.height
				);
		}
	} while(wan==0);
	
	if (wan!=dmx_screen_count) {
		mpiglutAbort("Error!  DMXGetWindowAttributes returned only %d windows, but there are %d screens!\n",
			wan,dmx_screen_count);
	}
	
	/* Check if DMX has a window on this screen */
	guiSendSync(MPIGLUT_DMX_PARENT_TAG);
	for (w=0;w<dmx_screen_count;w++) {
		int winNum=wa[w].window;
		if (wa[w].screen!=w)
			mpiglutAbort("Error!  DMXGetWindowAttributes entry %d is for screen %d!\n",
				w,wa[w].screen);
		guiSendData(&winNum,sizeof(winNum), "dmxFrontendCreateWindow sending off window ID");
	}
	guiSendSync(MPIGLUT_DMX_PARENT_TAG);
	free(wa);
	trace(10,"dmxFrontendCreateWindow complete\n");
}

INTERNAL Window dmxBackendCreateWindow(mpiglut_state_t *s,int parent,const mpiglut_rect *size)
{
	Window dmx_parent=0; /* DMX parent window */
	if (!dmx_enabled) return 0;
	trace(10,"dmxBackendCreateWindow starting (parent %d, %d DMX screens)\n",parent,dmx_screen_count);

// Receive new window ID from frontend
	if (s->rank==s->bossRank) 
	{ /* backend boss node: receive from frontend, send to minions */
		int scr;
		guiRecvSync(MPIGLUT_DMX_PARENT_TAG);
		for (scr=0;scr<dmx_screen_count;scr++) {
			int winNum=0;
			guiRecvData(&winNum,sizeof(winNum), "dmxBackendCreateWindow receiving window number");
			if (scr==s->rank) /* that's my window */
				dmx_parent=(Window)winNum;
			else if (scr<s->size) /* send window off to a minion */
				MPI_Send(&winNum,1,MPI_INT,scr, MPIGLUT_DMX_PARENT_TAG, s->comm);
		}
		guiRecvSync(MPIGLUT_DMX_PARENT_TAG);
	} 
	else /* backend minion node: receive from boss */ 
	{
		int winNum;  MPI_Status sts;
		MPI_Recv(&winNum,1,MPI_INT, s->bossRank,MPIGLUT_DMX_PARENT_TAG,s->comm,&sts);
		dmx_parent=(Window)winNum; /* an X11 "Window" is an "XID" is a "CARD32" is a 32-bit int. */
	}
	
	trace(10,"dmxBackendCreateWindow complete: GLUT parent=%d becomes X parent %x\n",parent,dmx_parent);
	//return 0; // for testing: don't bother with DMX parent window crap
	return dmx_parent;
	if (parent==0)  /* Only top-level windows need to have their DMX parent set */
		return dmx_parent;
	else /* parent!=0, so a child window--it's already got a parent. */
		return 0;
}
#endif

/********************************* (gui) GUI front-end receive code ***************************
  This chunk of code receives GLUT user-input events from the user-visible frontend GUI, 
  and forwards them all to the "guiSendEvent" routine, which sends them to the MPI backends.  
  
  The front end's job is to manage the user's interaction with a set of blank event-handling windows.
  It *doesn't* do any drawing.  It *doesn't* interact with user code.  All that happens on the backends.
*/

/* This event contains the last-known values for all GUI calls. */
INTERNAL mpiglut_event_t guiLast;

/** Set r, which was in global coordinates, to live in parent's coordinates */
INTERNAL void guiParentFromGlobal(int parent,mpiglut_rect *r)
{
	mpiglut_window_t *w;
	if (parent==0) return; /* no parent--already *in* global coordinates! */
	w=mpiglutWindow(parent);
	// Subtract off parent window's global coordinates
	mpiglut_rect_shift(-w->window.lo_x,-w->window.lo_y,r);
}

/** Set r, which was in global coordinates, to live in parent's coordinates */
INTERNAL void guiGlobalFromParent(int parent,mpiglut_rect *r)
{
	mpiglut_window_t *w;
	if (parent==0) return; /* no parent--already *in* global coordinates! */
	w=mpiglutWindow(parent);
	// Add in parent window's global coordinates
	mpiglut_rect_shift(+w->window.lo_x,+w->window.lo_y,r);
}
/** Return the global-coordinates bounds of the current window */
INTERNAL mpiglut_rect guiRectFromWindow(mpiglut_window_t *w) {
	mpiglut_rect r=mpiglut_rect_new(glutGet(GLUT_WINDOW_X),glutGet(GLUT_WINDOW_Y),
		glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
	/* glut returns parent coordinates.  shift to global coordinates */
	guiGlobalFromParent(w->parent,&r);
	return r;
}

INTERNAL void guiSetupEvent(mpiglut_event_t *e,int type) {
	mpiglut_window_t *w=mpiglutGetWindow();
	mpiglutMakeEvent(e,type);
	e->window=w->id;
	e->device=0;
	e->x=guiLast.x; e->y=guiLast.y;
	e->mousestate=guiLast.mousestate;
	e->modifiers=guiLast.modifiers;
	e->win=guiRectFromWindow(w);
	e->items=0;
	e->bytes=0;
	mpiglutTrace(15,"			guiSetupEvent(w=%d) t=%d, mod %x, win G "RECT_FMT"\n",e->window,e->type,e->modifiers,RECT_ARGS(e->win));
}

/* Return true if the current window's size has changed */
INTERNAL int guiWindowChanged(void) {
	mpiglut_window_t *w=mpiglutGetWindow();
	mpiglut_rect r=guiRectFromWindow(w);
	if (!mpiglut_rect_equal(&r,&w->window)) {
		w->window=r;
		return 1;
	} else {return 0;}
}
/* Check window sizes--if they've changed, send off a WINDOWCHANGE event. 
  This periodic polling seems to be the only way to detect window dragging...
*/
INTERNAL void guiWindowCheck(void)
{
	int s,w=glutGetWindow();
	for (s=0;s<n_window_slots;s++)
	if (window_slots[s]) {
		int id=window_slots[s]->id;
		glutSetWindow(id);
		if (guiWindowChanged()) 
		{ /* The current window has changed size-- send off a WINDOWCHANGE event, and check the window's children */
			mpiglut_event_t e;
			guiSetupEvent(&e,MPIGLUT_EVENT_WINDOWCHANGE);
			mpiglutTrace(20,"guiWindowCheck(w=%d) window size changed to G "RECT_FMT"\n",id,RECT_ARGS(e.win));
			guiSendEvent(&e);
		}
	}
	if (w!=0) glutSetWindow(w);
}

INTERNAL void guiMouseFn(int button,int state,int x,int y) {
	if (state==0) /* going down */ guiLast.mousestate|=(1<<button); 
	else /* going up */ guiLast.mousestate&=~(1<<button); 
	guiLast.x=x; guiLast.y=y;
	guiLast.modifiers=glutGetModifiers();
	mpiglut_event_t e; guiSetupEvent(&e,MPIGLUT_EVENT_MOUSEBUTTON);
	e.button=button; e.value=!state;
	guiSendEvent(&e);
}
INTERNAL void guiMotionFn(int x,int y) {
	guiLast.x=x; guiLast.y=y;
	mpiglut_event_t e; guiSetupEvent(&e,MPIGLUT_EVENT_MOUSEMOTION);
	guiSendEvent(&e);
}
INTERNAL void guiKeyboardFn(unsigned char key,int x,int y) {
	guiLast.x=x; guiLast.y=y;
	guiLast.modifiers=glutGetModifiers();
	mpiglut_event_t e; guiSetupEvent(&e,MPIGLUT_EVENT_KEYBOARD);
	e.button=key; e.value=1;
	guiSendEvent(&e);
}
INTERNAL void guiKeyboardUpFn(unsigned char key,int x,int y) {
	guiLast.x=x; guiLast.y=y;
	guiLast.modifiers=glutGetModifiers();
	mpiglut_event_t e; guiSetupEvent(&e,MPIGLUT_EVENT_KEYBOARD);
	e.button=key; e.value=0;
	guiSendEvent(&e);
}
INTERNAL void guiSpecialFn(int key,int x,int y) {
	guiLast.x=x; guiLast.y=y;
	guiLast.modifiers=glutGetModifiers();
	mpiglut_event_t e; guiSetupEvent(&e,MPIGLUT_EVENT_KEYBOARD);
	e.button=MPIGLUT_SPECIAL_START+key; e.value=1;
	guiSendEvent(&e);
}
INTERNAL void guiSpecialUpFn(int key,int x,int y) {
	guiLast.x=x; guiLast.y=y;
	guiLast.modifiers=glutGetModifiers();
	mpiglut_event_t e; guiSetupEvent(&e,MPIGLUT_EVENT_KEYBOARD);
	e.button=MPIGLUT_SPECIAL_START+key; e.value=0;
	guiSendEvent(&e);
}
INTERNAL void guiEnterFn(int state) {
	mpiglutTrace(10,"fg enter(w=%d) %d\n",glutGetWindow(),state);
	mpiglut_event_t e; guiSetupEvent(&e,MPIGLUT_EVENT_ENTER);
	e.value=state;
	guiSendEvent(&e);
}
INTERNAL void guiReshapeFn(int w,int h) {
	mpiglutTrace(10,"fg reshape(w=%d) %dx%d\n",glutGetWindow(),w,h);
	mpiglut_event_t e; guiSetupEvent(&e,MPIGLUT_EVENT_WINDOWCHANGE);
	guiSendEvent(&e);
}
INTERNAL void guiDisplayFn(void) {
	/* 
	  When everything's working, nothing is needed here--
	     the backends should display right on top of the frontend. 
	  Show some simple info anyway, for debugging...
	*/
	glDrawBuffer(GL_FRONT);
	glClearColor(1.0,0.5,0.5,0.0);
	glClear(GL_COLOR_BUFFER_BIT);
	glColor3f(0,0,0);
	glRasterPos3f(0,0,0); // 30,30,0);
	const char *str="mpiglut frontend display (are backends working?)";
	while (*str!=0) glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_10,*str++);
}

INTERNAL void guiAddFuncs(void) {
	glutMouseFunc(guiMouseFn);
	glutMotionFunc(guiMotionFn);
	glutPassiveMotionFunc(guiMotionFn);
	glutKeyboardFunc(guiKeyboardFn);
	glutKeyboardUpFunc(guiKeyboardUpFn);
	glutSpecialFunc(guiSpecialFn);
	glutSpecialUpFunc(guiSpecialUpFn);
	glutEntryFunc(guiEnterFn);
	glutReshapeFunc(guiReshapeFn);
	glutDisplayFunc(guiDisplayFn);
}

/** The frontend's idle function mostly waits for events from the backend */
INTERNAL void guiIdleFn(void) {
	guiWindowCheck();
	mpiglut_event_t e;
	if (!guiRecvEvent(&e)) { /* nothing to do */
		porthread_yield(1);
	}
	else /* we have an event from the network */
	switch (e.type) {
	case MPIGLUT_EVENT_WINDOWOPEN: { /* backend wants us to create a window */
		mpiglut_window_t *win=0;
		int parent=e.window, w;
		mpiglutTrace(10,"message from backend: WINDOWOPEN(parent=%d), G "RECT_FMT"\n",
			e.window, RECT_ARGS(e.win));
		mpiglutFrontendWindows++; 
		if (parent==0) { /* new top-level window */
			glutInitWindowPosition(e.win.lo_x,e.win.lo_y);
			glutInitWindowSize(mpiglut_rect_width(&e.win),
				mpiglut_rect_height(&e.win));
			w=glutCreateWindow("MPIglut Frontend Placeholder Window");
		} else { /* new subwindow */
			/* translate request from global to parent coordinates */
			guiParentFromGlobal(parent,&e.win);
			w=glutCreateSubWindow(parent,e.win.lo_x,e.win.lo_y,
				mpiglut_rect_width(&e.win),
				mpiglut_rect_height(&e.win));
		}
		glutSetWindow(w);
		guiAddFuncs(); /* make sure we receive events for this new window */
		win=mpiglutWindow(w);
		win->id=w;
		win->parent=parent;
#if USE_DMX
		dmxFrontendCreateWindow(w,parent,(Window)glutGetPointer(GLUT_WINDOW_HANDLE));
#endif
		// win->window=guiRectFromWindow(win);
		guiWindowCheck(); /* send WINDOWCHANGED event back to backend */
		mpiglutTrace(10,"   WINDOWOPEN handling complete, opened window %d, parent %d, at "RECT_FMT"\n",
			w,parent, RECT_ARGS(e.win));
		/* FIXME: hook up DMX subwindows to new backend windows */
		} break;
	case MPIGLUT_EVENT_WINDOWCHANGE: { /* backend requested a change in window positions */
		mpiglut_window_t *w=mpiglutWindow(e.window);
		mpiglutTrace(10,"message from backend: WINDOWCHANGE(w=%d), G "RECT_FMT"\n",
			e.window, RECT_ARGS(e.win));
		guiParentFromGlobal(w->parent,&e.win);
		glutSetWindow(e.window);
		glutPositionWindow(e.win.lo_x,e.win.lo_y);
		glutReshapeWindow(mpiglut_rect_width(&e.win),mpiglut_rect_height(&e.win));
		w->window=guiRectFromWindow(w);
		} break;
	case MPIGLUT_EVENT_WINDOWTITLE: { /* backend wants to change a window title */
		glutSetWindow(e.window);
		char *title=(char *)&e.MPIGLUT_EVENT_TITLE_START_FIELD;
		title[sizeof(e)-offsetof(mpiglut_event_t,MPIGLUT_EVENT_TITLE_START_FIELD)-1]=0; /* nul-terminate string */
		mpiglutTrace(10,"message from backend: WINDOWTITLE(w=%d), '%s'\n",
			e.window, title);
		switch (e.value) {
		case 0: glutSetWindowTitle(title); break;
		case 1: glutSetIconTitle(title); break;
		default: mpiglutAbort("mpiglut guiIdleFn> Unknown window title value %d!",e.value);
		}
		} break;
	case MPIGLUT_EVENT_WINDOWVIZ: { /* backend wants to change window visibility */
		mpiglutTrace(10,"message from backend: WINDOWVIZ(w=%d), value %d\n",
			e.window, e.value);
		glutSetWindow(e.window);
		switch (e.value) {
		case 0: glutShowWindow(); break;
		case 1: glutHideWindow(); break;
		case 2: glutIconifyWindow(); break;
		case 3: { /* destroying a window */
			mpiglutDestroyWindowSlot(e.window);
			glutDestroyWindow(e.window); 
			mpiglutFrontendWindows--; 
			break;
		}
		default: mpiglutAbort("mpiglut guiIdleFn> Unknown window visibility value %d!",e.value);
		}
		} break;
	default:
		mpiglutAbort("mpiglut guiIdleFn> Frontend received unexpected event type %d!\n",e.type);
	};
}

/** Processes events during the entire existence of the frontend */
INTERNAL void guiFrontendRun(int* pargc, char** argv)
{
	mpiglutTrace(1,"guiFrontendRun> ----------- calling glutInit with DISPLAY='%s'\n",getenv("DISPLAY"));
	glutInit(pargc,argv);
	glutIdleFunc(guiIdleFn);

	/* SILLY: we've got to wait until a window gets created before entering glutMainLoop,
	  or else the frontend's glutMainLoop will instantly exit! 
	*/
	while (mpiglutFrontendWindows==0)
		guiIdleFn();

	mpiglutTrace(1,"guiFrontendRun> ------------- entering main event loop\n",0);
	glutMainLoop();
	mpiglutTrace(1,"guiFrontendRun> ------------- finished with GLUT event loop.  Exiting.\n",0);
	exit(0);
}
